/* packet-mate.c
 * Routines for the mate Facility's Pseudo-Protocol dissection
 *
 * Copyright 2004, Luis E. Garcia Ontanon <lego@webflies.org>
 *
 * $Id: packet-mate.c,v 1.1 2004/08/24 23:28:41 lego Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**************************************************************************
 * This is the pseudo protocol dissector for the mate module.          ***
 * It is intended for this to be just the user interface to the module. ***
 **************************************************************************/

#include "plugins/plugin_api.h"
#include "plugins/plugin_api_defs.h"


#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "moduleinfo.h"

#include <stdio.h>
#include <string.h>
#include <gmodule.h>
#include <epan/packet.h>
#include <epan/strutil.h>
#include <epan/prefs.h>
#include "mate_util.h"
#include "mate.h"

#ifndef ENABLE_STATIC
G_MODULE_EXPORT const gchar version[] = VERSION;
#endif 


static int proto_mate = -1;
static gint ett_mate = -1;
static int hf_mate = -1;
static int hf_mate_leg = -1;
static int hf_mate_leg_attr = -1;
static int hf_mate_leg_pdu = -1;
static int hf_mate_leg_num_pdus = -1;
static int hf_mate_leg_start_time = -1;
static int hf_mate_leg_stop_time = -1;
static int hf_mate_leg_last_time = -1;
static int hf_mate_ses_num_legs = -1;
static int hf_mate_ses = -1;
static int hf_mate_pdu = -1;
static int hf_mate_pdu_rel_time = -1;


static char* pref_mate_config_filename = "config.mate";

void init_mate_dissector(void) {
	
	initialize_mate( pref_mate_config_filename );
	
}

void avpl_tree(AVPL* avpl,proto_tree* tree,tvbuff_t *tvb, int hfi) {
	AVPN* c;
	guint8* s;
	
	for ( c = avpl->null.next; c->avp; c = c->next) {
		
		s = avp_to_str(c->avp);
		
		if (hfi) {
			proto_tree_add_string_format(tree,hfi,tvb,0,0,s,"%s",s);
		} else {
			proto_tree_add_text(tree,tvb,0,0,"%s",s);
		}
		
		g_free(s);
	}
}

void mate_leg_tree(proto_tree* pdu_tree, tvbuff_t *tvb, mate_leg* leg);

void mate_ses_tree(proto_tree* leg_tree, tvbuff_t *tvb, mate_ses* ses) {
	proto_item *ses_item;
	proto_tree *ses_tree;
	proto_item *ses_leg_item;
	proto_tree *ses_leg_tree;
	proto_item *ses_attr_item;
	proto_tree *ses_attr_tree;
	mate_leg* ses_legs;
#ifdef _MATE_DEBUGGING
	proto_item* ses_key_item;
	proto_tree* ses_key_tree;
	guint i;
#endif
	
	ses_item = proto_tree_add_string(leg_tree,
									 hf_mate_ses,
									 tvb,0,0,
									 ses->id);
	ses_tree = proto_item_add_subtree(ses_item,ett_mate);
	
#ifdef _MATE_DEBUGGING
	if (ses->ses_keys->len) {
		ses_key_item = proto_tree_add_text(ses_tree,tvb,0,0,"Ses Key");
		
		ses_key_tree = proto_item_add_subtree(ses_key_item,ett_mate);
		
		for (i = 0 ; i < ses->ses_keys->len; i++) {
			proto_tree_add_text(ses_key_tree,tvb,0,0,"Key: %s",(guint8*) g_ptr_array_index(ses->ses_keys,i));
		}
	}
#endif
		
	ses_attr_item = proto_tree_add_text(ses_tree,tvb,0,0,"Ses Attributes");
	ses_attr_tree = proto_item_add_subtree(ses_attr_item, ett_mate);

	avpl_tree(ses->avpl,ses_attr_tree,tvb,hf_mate);
	
	ses_leg_item = proto_tree_add_uint_format(ses_tree, hf_mate_ses_num_legs, tvb, 0, 0,ses->num_of_legs, "Ses Legs (%u)", ses->num_of_legs);
	ses_leg_tree = proto_item_add_subtree(ses_leg_item, ett_mate);
	
	for (ses_legs = ses->legs; ses_legs; ses_legs = ses_legs->next) {
		mate_leg_tree(ses_leg_tree, tvb, ses_legs);
	}
	
}

void mate_leg_tree(proto_tree* pdu_tree, tvbuff_t *tvb, mate_leg* leg) {
	proto_item *leg_item;
	proto_tree *leg_time_tree;
	proto_item *leg_time_item;
	proto_tree *leg_tree;
	proto_item *leg_pdu_item;
	proto_tree *leg_pdu_tree;
	proto_item *leg_attr_item;
	proto_tree *leg_attr_tree;
	mate_pdu* leg_pdus;
	float  rel_time;
	float  leg_time;

	leg_item = proto_tree_add_string(pdu_tree,hf_mate_leg,tvb,0,0,leg->id);
	leg_tree = proto_item_add_subtree(leg_item, ett_mate);
	
#ifdef _MATE_DEBUGGING
	proto_tree_add_text(leg_tree,tvb,0,0,"Leg Key: %s",leg->leg_key);
#endif
	
	leg_attr_item = proto_tree_add_text(leg_tree,tvb,0,0,"Leg Attributes");
	leg_attr_tree = proto_item_add_subtree(leg_attr_item, ett_mate);
	
	avpl_tree(leg->avpl,leg_attr_tree,tvb,hf_mate_leg_attr);
	
		
	leg_time_item = proto_tree_add_text(leg_tree,tvb,0,0,"Leg Times");
	leg_time_tree = proto_item_add_subtree(leg_time_item, ett_mate);
	
	proto_tree_add_float(leg_time_tree, hf_mate_leg_start_time, tvb, 0, 0, leg->start_time);
				
	if (leg->released) { 
		proto_tree_add_float(leg_time_tree, hf_mate_leg_stop_time, tvb, 0, 0, leg->release_time);
		if (leg->release_time != leg->last_time) {
			proto_tree_add_float(leg_time_tree, hf_mate_leg_last_time, tvb, 0, 0, leg->last_time); 
		}
	} else {
		proto_tree_add_float(leg_time_tree, hf_mate_leg_last_time, tvb, 0, 0, leg->last_time); 
	}		

	rel_time = leg_time = leg->start_time;

	leg_pdu_item = proto_tree_add_uint_format(leg_tree, hf_mate_leg_num_pdus, tvb, 0, 0,leg->num_of_pdus, "Leg PDUs (%u)", leg->num_of_pdus);
	leg_pdu_tree = proto_item_add_subtree(leg_pdu_item, ett_mate);
	
	for (leg_pdus = leg->pdus; leg_pdus; leg_pdus = leg_pdus->next) {
		if (leg_pdus->is_start) {
			proto_tree_add_uint_format(leg_pdu_tree,hf_mate_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "Start PDU: in frame %i",
									   leg_pdus->framenum);
		} else if (leg_pdus->is_stop) {
			proto_tree_add_uint_format(leg_pdu_tree,hf_mate_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "Stop PDU: in frame %i (%f : %f)",
									   leg_pdus->framenum,
									   leg_pdus->rel_time,
									   leg_pdus->rel_time-rel_time);
			
		} else if (leg_pdus->after_release) {
			proto_tree_add_uint_format(leg_pdu_tree,hf_mate_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "After stop PDU: in frame %i (%f : %f)",
									   leg_pdus->framenum,
									   leg_pdus->rel_time,
									   leg_pdus->rel_time-rel_time);
		} else {
			proto_tree_add_uint_format(leg_pdu_tree,hf_mate_leg_pdu,
									   tvb,0,0,leg_pdus->framenum,
									   "PDU: in frame %i (%f : %f)",
									   leg_pdus->framenum,
									   leg_pdus->rel_time,
									   leg_pdus->rel_time-rel_time);
		}
		
		rel_time = leg_pdus->rel_time;
		
	}

	/* if ( leg->ses ) mate_ses_tree(leg_tree,tvb,leg->ses); */
}


void mate_pdu_tree(mate_pdu *pdu, tvbuff_t *tvb, proto_tree* tree) {
	proto_item *pdu_item;
	proto_tree *pdu_tree;
	proto_item *pdu_attr_item;
	proto_tree *pdu_attr_tree;
	
	guint32 len;
	
	if ( ! pdu ) return;
	
	len = pdu->end - pdu->start;
	pdu_item = proto_tree_add_text(tree,tvb,pdu->start,len,"%s",pdu->id);
	pdu_tree = proto_item_add_subtree(pdu_item, ett_mate);
	proto_tree_add_float(pdu_tree, hf_mate_pdu_rel_time, tvb, 0, 0, pdu->rel_time);		

	if (pdu->leg) {
		if (pdu->leg->ses) {
			proto_tree_add_text(pdu_tree,tvb,0,0,"This Leg: %s",pdu->leg->id);
			mate_ses_tree(pdu_tree,tvb,pdu->leg->ses);
		} else {
			mate_leg_tree(pdu_tree,tvb,pdu->leg);
		}
	}
	
	if (pdu->avpl) {
		pdu_attr_item = proto_tree_add_text(pdu_tree,tvb,0,0,"PDU Attributes");
		pdu_attr_tree = proto_item_add_subtree(pdu_attr_item, ett_mate);
		avpl_tree(pdu->avpl,pdu_attr_tree,tvb,hf_mate_pdu);
	}
	
}

extern void mate_tree(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree) {
	mate_pdu* pdus;
	proto_item *mate_i;
	proto_tree *mate_t;
	
	if (!tree) return;
	
	pdus = mate_get_pdus(pinfo->fd->num);
	
	if (pdus) {
		/* TODO: colorize */
		
		mate_i = proto_tree_add_uint_format(tree,hf_mate_leg_pdu,
												tvb,0,0,
												pinfo->fd->num,"mate");
		
		mate_t = proto_item_add_subtree(mate_i, ett_mate);
		
		for ( ; pdus; pdus = pdus->next_in_frame) {
			mate_pdu_tree(pdus,tvb,mate_t);			
		}
	}
}

void
proto_reg_handoff_mate(void)
{
	;
}  


void
proto_register_mate(void)
{                 
	static hf_register_info hf[] = {
	{ &hf_mate, {"mate", "mate",
		FT_STRING, BASE_NONE, NULL, 0,
		"Interface to mate Facility",HFILL }},
	{ &hf_mate_leg, {"Leg", "mate.leg",
		FT_STRING, BASE_NONE, NULL, 0,
		"A mate Leg",HFILL }},
	{ &hf_mate_ses, {"Session", "mate.session",
		FT_STRING, BASE_NONE, NULL, 0,
		"A mate Session",HFILL }},
	{ &hf_mate_pdu, {"PDU Attribute", "mate.pdu.attr",
		FT_STRING, BASE_NONE, NULL, 0,
		"An Attribute of a PDU",HFILL }},
	{ &hf_mate_leg_attr, {"Leg Attribute", "mate.leg.attr",
		FT_STRING, BASE_NONE, NULL, 0,
		"An Attribute of a Leg",HFILL }},
	{ &hf_mate_leg_pdu, {"Leg's PDU", "mate.leg.pdu",
		FT_FRAMENUM, BASE_DEC, NULL, 0,
		"a PDU that belongss to this leg",HFILL }},
	{ &hf_mate_pdu_rel_time, {"seconds since start", "mate.leg.pdu.time",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Seconds passed since the start of the leg or capture if the pdu is unnassigned",HFILL }},
	{ &hf_mate_leg_start_time, {"start time", "mate.leg.start_time",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Seconds passed since the begining of caputre to the start of this leg",HFILL }},
	{ &hf_mate_leg_stop_time, {"leg time", "mate.leg.duration",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Duration in seconds from start to stop of this leg",HFILL }},
	{ &hf_mate_leg_last_time, {"last pdu", "mate.leg.excess_duration",
		FT_FLOAT, BASE_DEC, NULL, 0,
		"Time passed between the start of this leg and the last pdu assigned to it",HFILL }},
	{ &hf_mate_ses_num_legs, {"number of legs", "mate.ses.legs",
		FT_UINT32, BASE_DEC, NULL, 0,
		"Number of legs in this session",HFILL }},
	{ &hf_mate_leg_num_pdus, {"number of PDUs", "mate.leg.pdus",
		FT_UINT32, BASE_DEC, NULL, 0,
		"Number of PDUs in this Leg",HFILL }},
	};

	static gint *ett[] = {
		&ett_mate,
	};

	module_t *mate_module;

	proto_mate = proto_register_protocol("The mate Thing", "mate", "mate");

	proto_register_field_array(proto_mate, hf, array_length(hf));

	proto_register_subtree_array(ett, array_length(ett));
	
	register_dissector("mate",mate_tree,proto_mate);

	register_init_routine(init_mate_dissector);

	mate_module = prefs_register_protocol(proto_mate, proto_reg_handoff_mate);
	
	prefs_register_string_preference(mate_module, "config_filename",
									 "Configuration Filename",
									 "The name of the file containing the mate module's configuration",
									 &pref_mate_config_filename);
	
}

#ifndef ENABLE_STATIC
G_MODULE_EXPORT void plugin_init(plugin_address_table_t *pat _U_ ) {
	/* initialise the table of pointers needed in Win32 DLLs */
	plugin_address_table_init(pat);
	
	/* register the new protocol, protocol fields, and subtrees */
	if (proto_mate == -1) { /* execute protocol initialization only once */
		proto_register_mate();
	}
}

G_MODULE_EXPORT void plugin_reg_handoff(void)
{
	proto_reg_handoff_mate();
}
#endif

